/***********************************************************************

FILE
    time2.c  -  implements 2 timers clocked off IBM PC's single timer

ROUTINES
    chronos    -  timer interrupt service routine
    SetDTimer  -  set countdown timer
    ReadDTimer -  read countdown timer
    TimeUp     -  time up yet?
    InitTime   -  initialize this module

REMARKS
    This is a simple implementation of two software timers which works
    much like hardware timers.  This module is needed since the IBM PC
    has only one timer and we need two timers to implement PI control
    with pulse width modulation.

    The xignal package and alarm.c are used to setup the timer interrupt.

    This module is a precursor to the full-blown general implementation
    of multiple timers in software.  If you want to use the general
    multiple software timer implementation, you should examine this 
    module first as it is much easier to understand than the general 
    implementation.

    Each timer is associated with an "interrupt" service routine which
    is executed whenever the associated timer/counter times out.

    Because of the speed limitations of the IBM PC, timer resolution
    is set at 1 millisecond.  Hence, SetDTimer() time argument is in
    milliseconds.  However, there is no reason why real clock ticks
    cannot be used instead of milliseconds.
    
    Since this is a special implementation to satisfy the requirements
    of pulse width modulation, each soft-timer is implemented with a 
    count variable instead of using an array of counters as in the 
    general implementation.  However, the implementation is such that 
    the module is independent of any specific applications - you can use
    this module for any applications on the IBM PC that requires just 
    two timers.  This also incurs less overheads than the general 
    implementation.

    Because this program replaces interrupt vectors used by DOS/BIOS,
    we cannot simply exit when an error is encountered, but must restore
    the DOS/BIOS interrupts to their default state first.  This is done
    in the panic() function (in file error.c).

    You will need to link this module with alarm.obj and xignal.lib

LAST UPDATE
    25 January 1988
        use void types and modify chronos()

    Copyright(c) 1985-1988  D.M. Auslander and C.H. Tham

***********************************************************************/

/***********************************************************************
                            I M P O R T S
***********************************************************************/

#include <stdio.h>

#include "envir.h"          /* environment definitions */
#include "8259.h"           /* 8259 programming declarations */
#include "xignal.h"         /* declarations for xignal package */
#include "alarm.h"          /* declarations for alarm.c */


/***********************************************************************
                    P R I V A T E     D A T A
***********************************************************************/

static long mstick;                 /* milliseconds per 'tick' */
static long halftick;               /* half of mstick */

static long time0, time1;           /* time values for both timers */
static void (*isr0)(), (*isr1)();   /* pointers to timeout routines */
static void *argp0, *argp1;         /* routine argument pointers */
static int isf0, isf1;              /* routine in-service flags */
static int flag0, flag1;            /* flags to indicate timeout */


/***********************************************************************
                    E N T R Y    R O U T I N E S
***********************************************************************/

/*----------------------------------------------------------------------
PROCEDURE
    chronos  -  timer interrupt service routine

SYNOPSIS
    void chronos() - called when hardware timer interrupt

REMARKS
    The software timer counters are decremented.  If any of the counts
    are zero or less, the associated "interrupt" service routine is
    executed provided the other conditions (timer has been set, there
    is an interrupt service routine declared and that routine is not 
    currently in service) are satisfied.

    Note that no priority scheme is implemented. 

    The user is responsible for ensuring that the timeout routines do
    not take too long to execute.  If another interrupt arrives and a
    timeout routine is still not finished, the other timeout routine
    (if installed) will be executed instead (if it is not similarly
    occupied, of course).  This means that the first timeout routine
    will miss the interrupt.

    The latest modification (Jan 88) only allows the timeout routine
    to execute once; a SetDTimer() call has to be made again before
    it will respond to another timer interrupt.  This is to prevent
    the routine from being called at every timer interrupt if the
    associated timer has timed out.  In a way, this is rather like
    edge-triggerred interrupts that are reset by calling SetDTimer().

LAST UPDATE
    25 January 1988
        execute isr only once, must reset timer to execute again.
----------------------------------------------------------------------*/

static void chronos()
{

    if (time0 > 0L)
        --time0;

    if (time1 > 0L)
        --time1;

    disable();              /* disables CPU's interrupt response */

    seoi(TMRVEC);           /* send Specific EOI to 8259 */

    /*-----------------------------------------------------------------
        If the timer has timed out, and there is a timeout routine
        associated with the timer and if the timeout routine is not
        currently executing, invoke the timeout routine, passing to
        it a pointer to possible arguments.
    -----------------------------------------------------------------*/

    if ((time0 <= 0L) && (isr0 != (void(*)())NULL) && (isf0 == 0) && !flag0)
    {
        flag0 = 1;          /* timer has timed out */

        isf0 = 1;           /* routine in service */

        enable();           /* enable interrupts */

        (*isr0)(argp0);     /* execute service routine */

        disable();          /* disable interrupts */

        isf0 = 0;           /* routine done */
    }
    
    if ((time1 <= 0L) && (isr1 != (void(*)())NULL) && (isf1 == 0) && !flag1)
    {
        flag1 = 1;          /* timer has timed out */

        isf1 = 1;           /* routine in service */

        enable();           /* enable interrupts */

        (*isr1)(argp1);     /* execute service routine */

        disable();          /* disable interrupts */

        isf1 = 0;           /* routine done */
    }

    enable();               /* enable interrupts before exit */

}



/*----------------------------------------------------------------------
PROCEDURE
    SetTsr  -  set timeout service routine and argument pointer

SYNOPSIS
    void SetTsr(timer, isf, argp)
    int timer
    void (*isf)(), *argp;

PARAMETERS
    timer  -  timer number (0 or 1)
    isf    -  pointer to timeout service routine
    argp   -  pointer to argument storage

LAST UPDATE
    25 January 1988
        clear flag0 and flag1
----------------------------------------------------------------------*/

void SetTsr(timer, isf, argp)
int timer;
void (*isf)(), *argp;
{
    int istat;      /* interrupt status */


    if (timer == 0)
    {
        istat = disable();

        isr0 = isf;
        isf0 = 0;
        argp0 = argp;
        flag0 = 0;

        if (istat)
            enable();
    }
    else if (timer == 1)
    {
        istat = disable();

        isr1 = isf;
        isf1 = 0;
        argp1 = argp;
        flag1 = 0;

        if (istat)
            enable();
    }
    else
    {
        fputs("bad timer number in SetTsr", stderr);
    }

}



/*----------------------------------------------------------------------
PROCEDURE
    SETDTIMER  -  set count down timer

SYNOPSIS
    void SetDTimer(timer, ms)
    int timer;
    long ms;

PARAMETERS
    timer  -  timer number (0 or 1)
    ms     -  time in milliseconds

REMARKS
    The halftick addition is used for rounding to the nearest tick.

    Because time0 and time1 are of type long (4 bytes on most 8088 C
    compilers), assignment will not be atomic; hence it has to be done
    inside a critical section.  Calculations are done outside the
    critical section to minimize the time interrupts have to be
    disabled.

    The timeout service routine disable flag is only cleared if the
    new setting is greater than 0.

LAST UPDATE
    25 January 1988
        assignment inside critical section
----------------------------------------------------------------------*/

void SetDTimer(timer, ms)
int timer;
long ms;
{
    int istat;      /* interrupt status */
    long tmp;       /* temporary for tick count */
    int flag;       /* flag status */


    tmp = (ms + halftick) / mstick;
    flag = (ms > 0L) ? 0 : 1;
    
    switch (timer)
    {
        case 0:

            istat = disable();

            time0 = tmp;
            flag0 = flag;

            if (istat)
                enable();

            break;

        case 1:

            istat = disable();
                
            time1 = tmp;
            flag1 = flag;

            if (istat)
                enable();

            break;

        default:

            fputs("bad timer number in SetDTimer", stderr);
    }

}



/*---------------------------------------------------------------------
FUNCTION
    READDTIMER  -  read time remaining

SYNOPSIS
    long ReadDTimer(timer)
    int timer;

PARAMETER
    timer  -  timer select (0 or 1)

RETURNS
    time remaining in milliseconds, resolution is mstick.

LAST UPDATE
    25 January 1988
        assignment to variable t inside critical section
----------------------------------------------------------------------*/

long ReadDTimer(timer)
int timer;
{
    int istat;          /* interrupt status */
    long t;             /* timer value */

    
    switch(timer)
    {
        case 0:

            istat = disable();

            t = time0;

            if (istat)
                enable();

            break;

        case 1:

            istat = disable();

            t = time0;

            if (istat)
                enable();

            break;

        default:
            fputs("bad timer number in ReadDTimer", stderr);
    }

    return(t * mstick);
}



/*----------------------------------------------------------------------
FUNCTION
    TIMEUP  -  has timer run out?

SYNOPSIS
    int TimeUp(timer)
    int timer;

PARAMETER
    timer  -  timer select (0 or 1)

RETURNS
    1 if true, 0 otherwise

LAST UPDATE
    22 April 1985
----------------------------------------------------------------------*/

int TimeUp(timer)
{
    int flag;           /* TimeUp flag */


    switch (timer)
    {
        case 0:
            flag = flag0;
            break;
        case 1:
            flag = flag1;
            break;
        default:
            flag = 1;
            fputs("bad timer number in TimeUp", stderr);
    }

    return(flag);
}



/***********************************************************************
            I N I T I A L I Z A T I O N    R O U T I N E S
***********************************************************************/

/*----------------------------------------------------------------------
PROCEDURE
    INITTIME  -  initialize this module

SYNOPSIS
    InitTime(ms)
    long ms;

PARAMETER
    ms  -  milliseconds per tick of driving hardware timer

LAST UPDATE
    25 January 1988
        also setup timer hardware and interrupts
----------------------------------------------------------------------*/

void InitTime(ms)
long ms;
{

    mstick = ms;
    halftick = mstick / 2L;     /* for rounding purposes */

    time0 = time1 = 0;
    isf0  = isf1  = 0;
    isr0  = isr1  = (void (*)())NULL;
    argp0 = argp1 = (void *)NULL;
    flag0 = flag1 = 1;

    xignal(XIGTMR, chronos);    /* install chronos() as timer isr */

}


